# -*- coding: utf-8 -*-

import logging
import os.path
import sys

from ..common.periodic_task import PeriodicTask
from ..filesync.filepath import is_hidden_part_in_path

_logger = logging.getLogger(__name__)


def files_list_updater(container, container_path, on_new_files,
                       on_changed_files, on_deleted_files,
                       on_initial_files=None, last_known_list=None,
                       check_period=600):
    """Detect changes in the files of a container.

    Each time a file is added, modified or removed, the corresponding callback
    is called.

    Args:
        session
        on_new_files (callable):
        on_changed_files (callable):
        on_deleted_files (callable):
        on_initial_files (callable, optional):
        last_known_list (dict(str, str), optional): known file. the key is the
            file's name and the value is the md5 sum.
    Returns:
        PeriodicTask: a task who update the list (by calling the callbacks) at
            a regular interval. It must be started using `start()`, and
            stopped with `stop()`
    """

    def update_list(pt, last_known_list):
        first_call = not pt.context
        if not first_call:
            last_known_list = pt.context['known_list']

        new_files = []
        changed_files = []
        initial_files = []
        deleted_files = []

        new_known_list = {}

        list_files = container.list_files().result()

        for f in list_files:
            abs_f = f['name']

            if sys.platform in ['win32', 'cygwin', 'win64']:
                abs_f = abs_f.replace('\\', '/')

            abs_f = os.path.join(container_path, abs_f)

            if is_hidden_part_in_path(container_path, abs_f):
                continue

            new_known_list[f['name']] = f['hash']
            if f['name'] in last_known_list:
                if f['hash'] != last_known_list[f['name']]:
                    changed_files.append(f)
                elif first_call and on_initial_files:
                    initial_files.append(f)
            else:
                new_files.append(f)

        for key in last_known_list:
            if key not in new_known_list:
                deleted_files.append({'name': key,
                                      'hash': last_known_list[key]})

        if new_files:
            on_new_files(new_files)
        if changed_files:
            on_changed_files(changed_files)
        if first_call and initial_files:
            on_initial_files(initial_files)
        if deleted_files:
            on_deleted_files(deleted_files)

        pt.context['known_list'] = new_known_list

    return PeriodicTask('File list updater %s' % container.id, check_period,
                        update_list, last_known_list or [])
