
import logging
import threading

from .api import Container
from .common.periodic_task import PeriodicTask
from .common.signal import Signal
from .promise import reduce_coroutine
from .local_container import LocalContainer

_logger = logging.getLogger(__name__)


class DynamicContainerList(object):
    """List of containers that the user has access.

    This list is dynamically updated, and can be used to get the list of
    containers the user has access, in a thread-safe way. It has methods to
    detect changes (through Signals), and wraps container-related methods about
    creation and deletions.

    The dynamic list must be started (by start()) to fetch the initial
    containers, and to detect subsequent changes. First, when data about
    containers listed in the user profile are loaded, the `loaded` signal is
    fired, with the exact same list as user profile.
    After that, `container_added` and `container_removed` signals are called
    each time a container is added or removed from the list.

    The good way to keep an up-to-date copy of the list is to call `get_list()`
    and to connect to all signals while owning the lock.


    Note:
        All actions who can add, remove, or hide a container from the user must
        be called from this list (and not from the LocalContainer class) to
        ensures the list is always up-to-date.

        All events are called owning the dynamic list lock.

        It's impossible to known when the dynamic list will be loaded (due to
        possible network error). The `loaded` event should always be considered
        even after the `start()` method has been called.

    Warning:
        `loaded` signal can emit invalid (deleted) containers: such local
        container has no container attribute.
        `container_removed` signal will fire the same containers almost
        immediately.

    Examples:

    make a copy `_my_list` of the container list, and keep it up-to-date.

    >>> with dyn_list.lock:
    >>>   _my_list = dyn_list.get_list()
    >>>
    >>>   def _on_loaded(new_list):
    >>>     _my_list.clear()
    >>>     _my_list.extends(new_list)
    >>>
    >>>   dyn_list.loaded.connect(_on_loaded)
    >>>   dyn_list.container_added.connect(_my_list.append)
    >>>   dyn_list.container_removed.connect(_my_list.remove)

    Attributes:
        lock (Lock): public lock, available to perform atomic operation on
            container list.
        loaded (Signal[List[LocalContainer]]): fired when the initial
            containers (already known) has been loaded.
        container_added (Signal[LocalContainer, boolean]): fired when a new
            container is detected.
            The second argument, 'from_local', is set to:
             - True if the container has been created by `create()`
             - False if the information has been obtained from the server.
        container_removed (Signal[LocalContainer, boolean]): fired when a
            container is destroyed (or when it's no longer shared with the
            actual user).
            The second argument, 'from_local', is set to:
             - True if the container has been created by `remove()`
             - False if the information has been obtained from the server.
    """

    # time between two list fetches (in seconds)
    CHECK_PERIOD = 600

    def __init__(self, session, user_profile):
        """DynamicContainerList constructor

        Args:
            session (Session): session required to do network operations.
            user_profile (UserProfile): profile from which the initial known
                container list is taken.
        """
        self._fetch_task = PeriodicTask('Dynamic container list updater',
                                        self.CHECK_PERIOD,
                                        self._update_list)
        self._session = session

        # Used at first update, to link known container models with api data.
        # Will be set to None after that.
        self._model_dict = user_profile.get_all_containers()

        self.lock = threading.Lock()
        self._container_dict = {}  # type: Dict[str, LocalContainer]

        self.loaded = Signal()
        self.container_added = Signal()
        self.container_removed = Signal()

    @reduce_coroutine()
    def create_share(self, name, is_encrypted, members, local_path=None,
                     do_not_sync=False):
        """Create a new TeamShare container and add it to the list.

        Args:
            name (Text)
            is_encrypted (boolean)
            members (Dict[Text, Dict[str, bool]]): list of members with theirs
                permissions. Ex:
                {u'foo@bar.fr': {'read': True, 'write': False, 'admin': False}}
            local_path (Optional[Text])
            do_not_sync (Optional[bool])
        Returns:
            (LocalContainer, boolean): lc created. the second value is True if
                an error occurred during the add of members. (the container
                exists but has not all members in).
        """
        with self.lock:
            lc, add_member_error = yield LocalContainer.create_share(
                self._session,
                name,
                is_encrypted,
                members,
                local_path,
                do_not_sync)

            self._container_dict[lc.id] = lc

            yield lc, add_member_error

    @reduce_coroutine(safeguard=True)
    def delete(self, container):
        """Delete a container definitively.

        Args:
            container (LocalContainer)
        Returns:
            Promise[None]
        """
        with self.lock:
            if container.id not in self._container_dict:
                _logger.warning('Unable to delete container: Container "%s" is'
                                ' not present in the list',
                                container.name)
                return
            yield container.container.delete()
            del self._container_dict[container.id]
        self.container_removed.fire(container, from_local=True)

    @reduce_coroutine(safeguard=True)
    def remove_self_access(self, container, user):
        """Remove user access to a container.

        Args:
            container (LocalContainer)
            user (Text): user email. It MUST be the current session owner.
        Returns:
            Promise[None]
        """
        with self.lock:
            if container.id not in self._container_dict:
                _logger.warning('Unable to quit container: Container "%s" is '
                                'not present ine the list',
                                container.name)
                return
            yield container.container.remove_member(user, is_self_quit=True)
            del self._container_dict[container.id]
        self.container_removed.fire(container, from_local=True)

    def get_list(self):
        """Get the list of containers visible by the user.

        Note:
            this method should always be called with the lock.
        Returns:
            List[LocalContainer]
        """
        return list(self._container_dict.values())

    def start(self):
        self._fetch_task.start()

    def stop(self):
        self._fetch_task.stop()

    def refresh(self):
        """Refresh the list immediately.

        Returns:
            Promise[List[LocalContainer]]: updated list.
        """
        return self._fetch_task.apply_now()

    def _update_list(self, _):
        with self.lock:
            container_list = Container.list(self._session).result()
            fetched_container_dict = {c.id: c for c in container_list}

            if self._model_dict is not None:  # First update
                for container_id, model in self._model_dict.items():
                    # container may be None. In that case, it'll produce a
                    # removed event immediately after the loaded event.
                    container = fetched_container_dict.get(container_id)
                    lc = LocalContainer(model, container)
                    self._container_dict[container_id] = lc
                self._model_dict = None
                self.loaded.fire(self.get_list())

            known_id_list = set(self._container_dict)
            fetched_id_list = set(fetched_container_dict)

            for container_id in fetched_id_list.difference(known_id_list):
                container = fetched_container_dict[container_id]
                lc = LocalContainer.from_api_data(container)

                self._container_dict[container_id] = lc
                self.container_added.fire(lc, from_local=False)

            for container_id in known_id_list.difference(fetched_id_list):
                container = self._container_dict.pop(container_id)
                self.container_removed.fire(container, from_local=False)
            return self.get_list()  # return value of self.refresh()
