# -*- coding: utf-8 -*-

# code based on
#   http://stackoverflow.com/questions/13420436
#
# definition of the static field in the classes
# (_reg_clsid_, _reg_progid_, ...)
#   http://docs.activestate.com/activepython/2.6/pywin32/html/com/win32com/HTML/QuickStartServerCom.html
#
# more information about overlay handler (C++) and the three handler
# methods (GetOverlayInfo, GetPriority, IsMemberOf)
#   https://www.codeproject.com/Articles/7484/How-to-overlay-an-icon-over-existing-shell-objects
#
# and a lot of information about com object in python (overlay handler is a
# com object).  It seems to be THE reference book about this topic.
#   https://books.google.fr/books?id=fzUCGtyg0MMC&pg=PA213

from multiprocessing.connection import Client
import pythoncom
import winerror
from win32com.shell import shell, shellcon

# to be print in the terminal, this print must stay before the import of
# win32traceutil
print("WARNING, the explorer must be restarted before to use the new"
      " Overlay handler!")  # noqa

# the next import will redirect/collect the output, it can be retrieve by
# launching the following command in another terminal:
#   python -m win32traceutil
import win32traceutil  # noqa

REG_PATH = (r'Software\Microsoft\Windows\CurrentVersion\Explorer'
            r'\ShellIconOverlayIdentifiers')
REG_KEY = "IconOverlayTest"


class IconOverlay(object):
    _public_methods_ = ['GetOverlayInfo', 'GetPriority', 'IsMemberOf']
    _com_interfaces_ = [shell.IID_IShellIconOverlayIdentifier,
                        pythoncom.IID_IDispatch]
    _reg_desc_ = 'Icon Overlay COM Server'

    def __init__(self, *_args, **_kwargs):
        self._icon_path = None

    def GetOverlayInfo(self):
        """
            This method is called ONLY once at the file explorer startup.
            (so at the early startup stage of windows...)
            It won't be called again, so the information returned by this
            method must be valid or the handler will never be able to work.
        """
        return self.get_icon(), 0, shellcon.ISIOI_ICONFILE

    def GetPriority(self):
        """
            Define the priority in front of what ? other handlers?

            Same as GetOverlayInfo, this method is called only once at
            file explorer startup.

            It must return an integer value between 0 and 100 (value included)
            0 is the higher, 100 is the lower
        """
        return 0

    def IsMemberOf(self, path, _attrs):
        """
            This method is called by the windows explorer to know if an emblem
            should be applied on the icon of the provided path.

            The return values allowed are winerror.S_OK (True) and
            winerror.E_FAIL (False)

            If bajoo is not running, it won't be possible to contact it and so
            this method will return False.
        """

        con = None
        try:
            con = Client(r'\\.\pipe\bajoo.overlay.pipe')
            con.send([path, self.get_handler_type()])
            result = con.recv()
        except:
            result = False
        finally:
            con and con.close()

        if result:
            return winerror.S_OK
        return winerror.E_FAIL

    def get_icon(self):
        """
            this method must return a valid icon path
        """
        pass

    def get_handler_type(self):
        """
            this method must return a key string known by the emblem
            manager.
        """
        pass


class IconOverlaySynced(IconOverlay):
    _reg_progid_ = r' bajoo.overlay.synced'
    _reg_clsid_ = '{64A9395C-BB13-4375-9D90-6B7FD3175EAF}'

    def get_icon(self):
        return "C:\\Users\\test\\Downloads\\1.ico"  # TODO set valid path

    def get_handler_type(self):
        return 'synced'


class IconOverlayNotSynced(IconOverlay):
    _reg_progid_ = r' bajoo.overlay.notsynced'
    _reg_clsid_ = '{9B138D6C-DD78-44D7-B72F-6434F2FBF895}'

    def get_icon(self):
        return "C:\\Users\\test\\Downloads\\2.ico"  # TODO set valid path

    def get_handler_type(self):
        return 'not_synced'

# ***
# *** ADD NEW HANDLER HERE ***
# ***

# *** HOW TO ADD A NEW HANDLER ***
#   1) create a new class that inherit from IconOverlay
#   2) define a new value for the class field _reg_progid_
#   3) generate a new value for the class field _reg_clsid_
#       with pythoncom.CreateGuid()
#   4) change the icon path in getIcon (where are the icons on windows ?)
#   5) define a new value returned by the function get_handler_type.  This
#       value will be communicate to Bajoo when asking if a path must have an
#       emblem.
#   6) add the new class in the following list (classes)
#   7) exececute this python file
#       to have feedback on this step, execute 'python -m win32traceutil' in
#       an other terminal
#   8) reboot the computer, the explorer only loads handler at start up

classes = [IconOverlaySynced, IconOverlayNotSynced, ]


def dll_register_server():
    """
        This function is called by UseCommandLine when no unregistering flag
        is set.  (see main function at the end of this file)
    """

    import _winreg
    winreg = _winreg
    for view in [winreg.KEY_WOW64_64KEY, winreg.KEY_WOW64_32KEY]:
        for cls in classes:
            print("Registering %s ......." % cls._reg_desc_)
            with winreg.CreateKeyEx(winreg.HKEY_LOCAL_MACHINE,
                                    r"%s\%s" % (REG_PATH, cls._reg_progid_),
                                    0,
                                    winreg.KEY_ALL_ACCESS | view) as hkey:
                print("    %s" % cls)
                winreg.SetValueEx(hkey,
                                  None,
                                  0,
                                  winreg.REG_SZ,
                                  cls._reg_clsid_)
            print("Registration complete: %s" % cls._reg_desc_)


def dll_unregister_server():
    """
        This function is called by UseCommandLine when a unregistering flag
        is set.  '--unregister' or '--unregisterInfo'.  (see main function at
        the end of this file)
    """
    import _winreg
    winreg = _winreg
    for view in [winreg.KEY_WOW64_64KEY, winreg.KEY_WOW64_32KEY]:
        for cls in classes:
            print("Unregistering %s ......." % cls._reg_desc_)
            try:
                winreg.DeleteKeyEx(winreg.HKEY_LOCAL_MACHINE,
                                   r"%s\%s" % (REG_PATH, cls._reg_progid_),
                                   winreg.KEY_ALL_ACCESS | view)
            except WindowsError as err:
                if err.errno != 2:
                    raise
            print("Unregistration complete: %s" % cls._reg_desc_)


if __name__ == '__main__':
# src: https://books.google.fr/books?id=fzUCGtyg0MMC&pg=PA217&lpg=PA217  # noqa
# src: https://github.com/SublimeText/Pywin32/blob/master/lib/x32/win32com/server/register.py#L578  # noqa
    """
        UseCommandLine uses sys.argv to define the values in flags if
        not provided.
        Two extra argument are only extracted from sys.argv:
            --unregister(default: False): unregister the com server classes
            --unregisterInfo(default: False): another way to unregister,
                do not find difference

        Flags:
            * quiet (default: False): no output in success case
            * unattended (default: False): it will raise at the first error.
              If the application does not have admin privelege, it won't try to
              get them, it will raise.
            * debug (default: False): add extra information at output

        Remember, because of the win32traceutil import, the output does not go
        to stdout.  need to execute 'python -m win32traceutil' in an other
        terminal to retrieve the output and debug.
    """

    flags = {}
    # flags['quiet'] = True
    flags['debug'] = True  # TODO comment
    # flags['unattended'] = True
    flags['finalize_register'] = dll_register_server
    flags['finalize_unregister'] = dll_unregister_server

    from win32com.server import register
    register.UseCommandLine(*classes, **flags)
