# -*- coding: utf-8 -*-

# code inspired from
#   https://sourceforge.net/p/tortoisesvn/code/HEAD/tree/trunk/src/TortoiseOverlays/IconOverlay.cpp
#
# server com object (guid generating, registering, ...):
#   http://docs.activestate.com/activepython/2.6/pywin32/html/com/win32com/HTML/QuickStartServerCom.html
#
# client com object (server testing, ...):
#   http://docs.activestate.com/activepython/2.6/pywin32/html/com/win32com/HTML/QuickStartClientCom.html
#
# code about the shutdown
#   https://mail.python.org/pipermail/python-win32/2008-December/008481.html

import win32con
import win32api
import winerror
import win32security

from ntsecuritycon import TOKEN_ADJUST_PRIVILEGES, SE_SHUTDOWN_NAME, \
    SE_PRIVILEGE_ENABLED, TOKEN_QUERY


PARENT_KEY_PATH = (r'Software\Microsoft\Windows\CurrentVersion\Explorer'
                   r'\ShellIconOverlayIdentifiers')


def error_reverse_lookup(error_int):
    """
        this function looks into the winerror module to find the name
        associated to an error code.  Because the text provided with the
        error code is translated and in an unknown encoding, it is easier to
        retrieve the name in the module.
    """
    for name in dir(winerror):
        if error_int == getattr(winerror, name):
            return name
    return None


def _get_overlay_handlers(path=PARENT_KEY_PATH):
    result = {}
    try:
        hkey = win32api.RegOpenKeyEx(win32con.HKEY_LOCAL_MACHINE,
                                     path,
                                     0,
                                     win32con.KEY_READ)
    except win32api.error, details:
        print('failed to open: '+str(error_reverse_lookup(details[0])))
        return result

    try:
        for progid, reserved, class_, mtime in win32api.RegEnumKeyEx(hkey):
            result[progid] = (reserved, class_, mtime,)

    except win32api.error, details:
        print('failed to list: '+str(error_reverse_lookup(details[0])))

    try:
        win32api.RegCloseKey(hkey)
    except:
        print('failed to close: '+str(error_reverse_lookup(details[0])))

    return result


def list_handler_progid(path=PARENT_KEY_PATH):
    handlers = _get_overlay_handlers(path)

    for progid, (reserved, class_, mtime) in handlers.items():
        print("progid: "+str(progid)+", reserved: "+str(reserved) +
              ", class: "+str(class_)+", mtime: "+str(mtime))


def handler_count(path=PARENT_KEY_PATH):
    return len(_get_overlay_handlers(path))


def is_handler_enabled(handler_progid,
                       path=PARENT_KEY_PATH,
                       limit=11):
    # 15 overlay handlers are available but 4 are reserved to windows...
    # (but onedrive use 5...)
    handlers = _get_overlay_handlers(path)

    if handler_progid not in handlers:
        return False

    if len(handlers) < limit+1:
        return True

    # is the sorting the same as windows ?
    keys = sorted(handlers.keys())
    i = keys.index(handler_progid)
    return i < limit


def is_handler_registered(handler_progid, path=PARENT_KEY_PATH):
    complete_path = path + '\\' + handler_progid
    try:
        hkey = win32api.RegOpenKeyEx(win32con.HKEY_LOCAL_MACHINE,
                                     complete_path,
                                     0,
                                     win32con.KEY_READ)
        return True
    except win32api.error, details:
        if details[0] != winerror.CDERR_INITIALIZATION:  # do not exist
            print('error while open: '+str(error_reverse_lookup(details[0])))
    else:
        win32api.RegCloseKey(hkey)

    return False


def generate_win_guid():
    import pythoncom
    print(str(pythoncom.CreateGuid()))


def _adjust_privilege(priv, enable=True):
    # Get the process token.
    flags = TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY
    htoken = win32security.OpenProcessToken(win32api.GetCurrentProcess(),
                                            flags)
    # Get the ID for the system shutdown privilege.
    id = win32security.LookupPrivilegeValue(None, priv)
    # Now obtain the privilege for this process.
    # Create a list of the privileges to be added.
    if enable:
        newPrivileges = [(id, SE_PRIVILEGE_ENABLED)]
    else:
        newPrivileges = [(id, 0)]
    # and make the adjustment.
    win32security.AdjustTokenPrivileges(htoken, 0, newPrivileges)


def reboot_windows(message="reboot...",
                   timeout=30,
                   force=False,
                   reboot=True):
    """
        this method reboot the computer.

        WARNING: this do not ask confirmation to the user.
    """
    _adjust_privilege(SE_SHUTDOWN_NAME)
    try:
        win32api.InitiateSystemShutdown(None, message, timeout, force, reboot)
    finally:
        # Now we remove the privilege we just added.
        _adjust_privilege(SE_SHUTDOWN_NAME, 0)


def abort_windows_reboot():
    _adjust_privilege(SE_SHUTDOWN_NAME)
    try:
        win32api.AbortSystemShutdown(None)
    finally:
        # Now we remove the privilege we just added.
        _adjust_privilege(SE_SHUTDOWN_NAME, 0)
