
import logging
from ..common.signal import Signal

_logger = logging.getLogger(__name__)


class BaseController(object):
    """Base class for all GUI controllers

    Each GUI element is split in two part: the controller and the view.
    The controller is in charge of the logic and the data processing.


    ## Dependency injection of the view and import rules

    As there can be multiple view implementations for one controller, the view
    is created by dependency injection. The controller receive a view class
    as parameter, and has no need to known the real view implementation type.
    A controller must never import a view package.

    GUI elements can be imported from the module parent to this element
    (eg: `from .gui.windows import AboutWindow`). However, if the GUI element
    is a sibling (eg: another window), the target element must be imported by
    its full package name
    (eg: `from .gui.windows.about_window import AboutWindow`).


    ## Interactions between view and controller

    The controller can call any method of the view which is decorated by a
    @public_method decorator. The view can only call "action" controller
    methods: these methods are passed as parameter at view construction.
    If a controller must share private data structure or enumerations with its
    views, they should be defined in the controller file. Public data
    structures and enumerations should be defined in the `bajoo.gui.enums`
    package.

    Attributes:
        view (BaseView): associated view (created by the factory).
        destroyed (Signal[]): signal fired when the controller is about to be
            destroyed.
        _destroyed (Boolean): True if the controller has been destroyed (and
            shouldn't be used anymore); default to False.
    """

    def __init__(self, view_factory, *args, **kwargs):
        """Controller constructor

        Args:
            view_factory (Callable[..., BaseView]): factory of view.
            *args (Tuple[int]): arguments passed to the view factory
            **kwargs: named arguments passed to the view factory
        """
        self.view = view_factory.make(self, *args, **kwargs)
        self.destroyed = Signal()
        self._destroyed = False

    def destroy(self):
        """Destroy the controller and its view.

        It should be the last method called from the controller.
        """
        if not self._destroyed:
            self._destroyed = True
            self.destroyed.fire()
            self.view.destroy()
        else:
            _logger.warning('multiple tentative to destroy instance of %s' %
                            self.__class__.__name__)
