# -*- coding: utf-8 -*-

from ..promise import Deferred
from functools import partial


class BaseView(object):
    """Base class for all GUI view classes.

    This classes provides utilities to communicate in a safe way between views
    and controllers. There are two important points to take in consideration
    for view-controller communication: thread-safety and separation of concern.

    1. Thread safety

    GUI frameworks have limitations on how to call methods in a multi-threaded
    environment: all calls must be done from the same thread.
    On the other side, controller methods can be executed by arbitrary threads.

    2. Separation of concern

    Views are dedicated to the user interface, and should not have any detail
    of the controller. On the other side, controllers should be GUI-agnostic,
    and accepts wxPython views as well as GTK views. Both views and controller
    interfaces should be obvious to the reader.


    To do so, this classes implements some communication mechanisms.

    1. View construction

    Views should be created by using the `make()` static method. It's a
    wrapping of the real constructor, which ensures the view constructor will
    always be called from the GUI thread. It's also required to use actions.

    2. Public view methods

    View methods destined to be called by the controller must use the decorator
    `@public_method`. It will wrap the method result in a Promise, and always
    call the method in the GUI thread.
    Contrary to `make()`, calls are asynchronous.

    3. Actions

    An Action is a way to call a constructor method by the view in response to
    an event.
    Actions are callbacks passed by constructor to `make()`. The view
    constructor will "declare" them by calling `self.action('name')` and
    receiving an asynchronous callable.
    All actions must be provided by the constructor, and passing an unknown
    action to the views will raises an error.
    An Action callable always returns None: the view express the user's will to
    performs an action, and it's up to the controller to do something, or not.

    ####

    All subclasses must use `super().__init__()`. Theirs constructors must
    accept arbitrary keyword arguments, and pass them to the super() init.

    """

    def __init__(self, actions={}, **kwargs):
        """BaseView constructor. Takes care of action declarations.

        Args:
            actions (Dict[str, Callable[..., None]): named argument, list of
                callback associated to each action defined in cls. The argument
                is consumed.
            **kwargs: ignored
        """
        super(BaseView, self).__init__(**kwargs)
        self._actions = actions

    def action(self, name):
        """Convert an action received in a usable callable.

        Note:
            This method MUST be called only in the child constructor, and
            after the BaseView constructor call.

        Args:
            name (str): name of the action to convert into callable.
        Returns:
            Callable[..., None]
        """
        try:
            action = self._actions.pop(name)
        except KeyError:
            raise KeyError('action %s not found' % name)
        return partial(self.async_gui_exec, action)

    @classmethod
    def make(cls, *args, **kwargs):
        """Create a new instance of the class.

        This method must be called instead of the __init__() constructor, as it
        ensures the constructor is always called in the thread required by the
        GUI framework.

        Args:
            *args: argument passed to the constructor.
            **kwargs: named argument passed to the constructor.
        Returns:
            cls: Value returned by `cls(*args, **kwargs)`
        Raises:
            Exception: If actions passed by argument and these consumed by the
                view doesn't match, an exception is raised. Otherwise, anything
                raised by the call to `cls(*args, **kwargs)`.
        """

        def _call_ctor():
            view = cls(*args, **kwargs)
            if view._actions:
                raise Exception('Some action(s) passed in parameter are not '
                                'consumed: %s' % view._actions.keys())
            view._actions = None
            return view

        if cls.is_in_gui_thread():
            return _call_ctor()
        else:
            df = Deferred()
            cls.async_gui_exec(df.wrap_function, _call_ctor)
            return df.promise.result()

    @classmethod
    def public_method(cls, method):
        """Wrap a GUI method and make it an async, publicly-callable method.

        This decorator will execute the function only in the GUI thread.
        It will delay the execution, and wrap the result in a Promise.

        Notes:
            If the wrapped function returns a Promise, this Promise will be
            chained, in order to never have nested promises.

        Args:
            method (Callable[..., T]): method to decorate. Will always be
                called in the GUI context.
        Returns:
            Callable[..., Promise[T]]: new method. It calls the real method in
                an async way, and returns a Promise wrapping the result.
        """
        def wrapper(*args, **kwargs):
            df = Deferred()
            cls.async_gui_exec(df.wrap_function, method, *args, **kwargs)
            return df.promise
        return wrapper

    @staticmethod
    def is_in_gui_thread():
        """Indicate if the actual thread is the GUI thread.

        Returns:
            bool: True if the current thread is the GUI thread; otherwise False
        """
        raise NotImplementedError()

    @staticmethod
    def async_gui_exec(fn, *args, **kwargs):
        """Asynchronously execute a callable in the GUI thread.

        Args:
            fn (Callable): callable to execute
            *args: argument passed to the callable
            **kwargs: named argument passed to the callable
        """
        raise NotImplementedError()

    def destroy(self):
        """Destroy the view (free all object related to the view)

        No view method should be called after a destroy.
        """
        raise NotImplementedError()
