
import wx


class DynamicText(wx.StaticText):
    """Enhanced wx.StaticText, with better support of dynamic update.

    Instances of DynamicText should always be part of wx.Sizer with the
    wx.EXTEND flag.

    This class offers the exact same interface as wx.StaticText.
    The difference resides in internal handling of resize events and text
    modification.

    The text uses all the available space, and automatically add line jumps if
    required. In case of text change (by `SetLabel()`), the size is also
    updated.
    """

    def __init__(self, *args, **kwargs):
        self._last_wrap_size = None
        self._ongoing_resize = False
        self._unwrapped_text = kwargs.get('label')

        super(DynamicText, self).__init__(*args, **kwargs)
        self.Bind(wx.EVT_SIZE, self._on_size_event)

    def _layout(self):
        """call Layout() on self and all ancestors."""
        w = self
        while w:
            w.Layout()
            w = w.GetParent()

    def _on_size_event(self, event):
        """"Rework the text after a size change."""
        event.Skip()

        if self._ongoing_resize:
            return  # Avoid recursive call
        self._ongoing_resize = True
        try:
            if self._unwrapped_text is None:
                return  # text undefined
            if self.Size.width == 0 or self.Size.height == 0:
                return  # Avoid infinite loop with wxPython GTK2
            if self._last_wrap_size == self.Size:
                return  # exact same Wrap already done previously.

            # wrap size must be saved before the call to SetLabel():
            # SetLabel() updates self.Size without the layout constraints
            self._last_wrap_size = self.Size
            wx.StaticText.SetLabel(self, self._unwrapped_text)
            self.Wrap(self._last_wrap_size.width)
            self._layout()
        finally:
            self._ongoing_resize = False

    def GetLabel(self):
        return self._unwrapped_text

    def SetLabel(self, label):
        self._last_wrap_size = None
        self._unwrapped_text = label
        wx.StaticText.SetLabel(self, label)
        self.SetMinSize((self.BestSize.width, -1))
