
import wx
from wx.lib.newevent import NewCommandEvent

from ...common.i18n import N_
from ..base_wx_view import BaseWxView
from ..validator import ConfirmPasswordValidator, MinLengthValidator
from . import BaseForm


class PassphraseFormField(BaseForm):
    """A widget asking an optional new passphrase.

    It's composed of two password inputs to ask the passphrase and its
    confirmation, and a checkbox to disable the passphrase.

    If the checkbox is not checked (it is by default), passphrase fields are
    disabled.
    """

    SubmitEvent, EVT_SUBMIT = NewCommandEvent()

    fields = ['no_passphrase', 'passphrase', 'confirmation',
              'allow_save_on_disk']

    def __init__(self, parent, **kwargs):
        BaseForm.__init__(self, parent, auto_disable=True, **kwargs)
        self._view = PassphraseFormFieldView(self)

        self.validators = [
            self.FindWindow('passphrase_validator'),
            self.FindWindow('confirmation_validator')
        ]

        self.Bind(wx.EVT_CHECKBOX, self.apply_field_constraints)

    def apply_field_constraints(self, _event=None):
        use_passphrase = not self.FindWindow('no_passphrase').IsChecked()
        self.FindWindow('passphrase').Enable(use_passphrase)
        self.FindWindow('confirmation').Enable(use_passphrase)
        self.FindWindow('passphrase_validator').reset()
        self.FindWindow('confirmation_validator').reset()
        self.FindWindow('allow_save_on_disk').Enable(use_passphrase)

    def get_data(self):
        data = BaseForm.get_data(self)

        if data['no_passphrase']:
            data['passphrase'] = None
        del data['no_passphrase']
        del data['confirmation']
        return data

    def reset_form(self):
        self.apply_field_constraints()
        self.enable()

    def notify_lang_change(self):
        BaseForm.notify_lang_change(self)
        self._view.notify_lang_change()


class PassphraseFormFieldView(BaseWxView):
    """View of the RegisterForm"""

    def __init__(self, parent):
        BaseWxView.__init__(self, parent)

        encryption_txt = wx.StaticText(self.window)
        passphrase = wx.TextCtrl(self.window, name='passphrase',
                                 style=wx.TE_PASSWORD)
        passphrase_validator = MinLengthValidator(
            self.window, name='passphrase_validator',
            target=passphrase, min_length=8)
        confirmation = wx.TextCtrl(self.window, name='confirmation',
                                   style=wx.TE_PASSWORD)
        confirmation_validator = ConfirmPasswordValidator(
            self.window, name='confirmation_validator',
            target=confirmation, ref=passphrase)
        no_passphrase = wx.CheckBox(self.window, name='no_passphrase')

        allow_save_on_disk_checkbox = wx.CheckBox(
            self.window, name='allow_save_on_disk')

        self.register_many_i18n('SetLabel', {
            encryption_txt: N_("Your passphrase is known only by yourself. "
                               "It's used to encrypt your data.\n"
                               "It should contains at least 8 characters."
                               "You can use a real phrase to memorize it more "
                               "easily (ex: MyBajooAccount)."),
            no_passphrase: N_("Don't use encryption passphrase"),
            allow_save_on_disk_checkbox: N_('Memorize the passphrase'),
        })

        self.register_many_i18n('SetHint', {
            passphrase: N_('Passphrase'),
            confirmation: N_('passphrase confirmation')
        })

        passphrase_sizer = self.make_sizer(wx.HORIZONTAL, [
            passphrase, passphrase_validator
        ], outside_border=False, flag=wx.EXPAND, proportion=1)
        passphrase_sizer.GetItem(passphrase).Proportion = 2
        confirmation_sizer = self.make_sizer(wx.HORIZONTAL, [
            confirmation, confirmation_validator
        ], outside_border=False, flag=wx.EXPAND, proportion=1)
        confirmation_sizer.GetItem(confirmation).Proportion = 2

        encryption_form_sizer = self.make_sizer(wx.VERTICAL, [
            encryption_txt, passphrase_sizer, confirmation_sizer,
            allow_save_on_disk_checkbox, no_passphrase
        ], flag=wx.EXPAND, outside_border=False)
        self.window.SetSizer(encryption_form_sizer)
