# -*- coding: utf-8 -*-

import wx
from wx.lib.newevent import NewCommandEvent

from ...common.i18n import N_
from ...common.path import default_root_folder
from ..base_wx_view import BaseWxView
from ..form import BaseForm
from ..form.passphrase_form_field import PassphraseFormField
from ..validator import BaseValidator


class SetupConfigScreen(BaseForm):
    """Ask the user about settings needed to start the sync.

    Attributes:
        EVT_SUBMIT (Event Id): event emitted when the form is submitted.
    """

    SubmitEvent, EVT_SUBMIT = NewCommandEvent()

    fields = ['passphrase_widget', 'bajoo_folder']

    def __init__(self, parent):
        BaseForm.__init__(self, parent, auto_disable=True)
        self._view = SetupConfigScreenView(self)
        self.add_i18n_child(self._view)

        self.validators = [
            self.FindWindow('gpg_error'),
            self.FindWindow('root_folder_error'),
            self.FindWindow('passphrase_widget').validators
        ]

        self.FindWindow('bajoo_folder').SetPath(default_root_folder())

        self.Bind(wx.EVT_BUTTON, self.submit,
                  source=self.FindWindow('validate_btn'))

    def get_data(self):
        data = BaseForm.get_data(self)
        data.update(data.pop('passphrase_widget', {}))
        return data

    def reset_form(self, folder_setting, key_setting, root_folder_error=None,
                   gpg_error=None):

        root_folder_error_txt = self.FindWindow('root_folder_error')
        if root_folder_error:
            root_folder_error_txt.set_msg(root_folder_error)
        self.FindWindow('bajoo_folder_section').Show(folder_setting)

        if gpg_error:
            gpg_error_txt = self.FindWindow('gpg_error')
            gpg_error_txt.set_msg(gpg_error)
        self.FindWindow('passphrase_widget').apply_field_constraints()
        self.FindWindow('passphrase_widget').Show(key_setting)
        self.FindWindow('encryption_section').Show(
            bool(key_setting or gpg_error))

        self.enable()

        self._reset_validators(key_setting, folder_setting)
        self.GetTopLevelParent().Layout()

    def _reset_validators(self, key_setting, folder_setting):
        # Reset validators
        self.validators = []

        if key_setting:
            self.validators.append(self.FindWindow('gpg_error'))
            self.validators += self.FindWindow('passphrase_widget').validators

        if folder_setting:
            self.validators.append(self.FindWindow('root_folder_error'))


class SetupConfigScreenView(BaseWxView):
    def __init__(self, screen):
        BaseWxView.__init__(self, screen)

        self.set_frame_title(N_('Bajoo - Configuration'))

        validate_btn = wx.Button(self.window, id=wx.ID_OK, name='validate_btn')

        encryption_section = wx.StaticBox(self.window,
                                          name='encryption_section')
        gpg_error = BaseValidator(encryption_section, hide_if_valid=True,
                                  name='gpg_error')

        passphrase_widget = PassphraseFormField(encryption_section,
                                                name='passphrase_widget')

        bajoo_folder_section = wx.StaticBox(self.window,
                                            name='bajoo_folder_section')
        root_folder_error = BaseValidator(bajoo_folder_section,
                                          hide_if_valid=True,
                                          name='root_folder_error')

        bajoo_folder_label = wx.StaticText(bajoo_folder_section)
        folder_picker = wx.DirPickerCtrl(bajoo_folder_section,
                                         name='bajoo_folder',
                                         style=wx.DIRP_USE_TEXTCTRL)

        self.register_many_i18n('SetLabel', {
            encryption_section: N_('Encryption'),
            bajoo_folder_section: N_('Root folder Bajoo'),
            bajoo_folder_label: N_('Your Bajoo shares will be deposit here:'),
            validate_btn: N_('Validate')
        })
        self.add_i18n_child(passphrase_widget)

        encryption_sizer = wx.StaticBoxSizer(encryption_section, wx.VERTICAL)
        self.make_sizer(wx.VERTICAL, [gpg_error, passphrase_widget],
                        sizer=encryption_sizer, flag=wx.EXPAND)

        bajoo_folder_sizer = wx.StaticBoxSizer(bajoo_folder_section,
                                               wx.VERTICAL)

        self.make_sizer(wx.VERTICAL, [
            root_folder_error, [bajoo_folder_label, folder_picker]
        ], flag=wx.EXPAND, sizer=bajoo_folder_sizer)
        bajoo_folder_sizer.GetItem(folder_picker,
                                   recursive=True).Proportion = 1

        btn_sizer = wx.StdDialogButtonSizer()
        btn_sizer.AddButton(validate_btn)
        btn_sizer.Realize()

        sizer = self.make_sizer(wx.VERTICAL, [
            encryption_sizer, bajoo_folder_sizer, None, btn_sizer
        ], flag=wx.EXPAND)
        self.window.SetSizer(sizer)


def main():
    app = wx.App()
    win = wx.Frame(None)
    app.SetTopWindow(win)
    screen = SetupConfigScreen(win)
    screen.GetSizer().SetSizeHints(win)
    win.Show(True)
    app.MainLoop()


if __name__ == '__main__':
    main()
