
import logging

from ....common.i18n import _
from ....common.strings import err2unicode
from ....network.errors import HTTPUnauthorizedError
from ....promise import reduce_coroutine
from ..base_window_controller import BaseWindowController

_logger = logging.getLogger(__name__)


class AccountDeletionWindowController(BaseWindowController):
    """Confirmation window for account deletion.

    The window must clearly explain that the action is irreversible. The user
    can confirms (and its account will be deleted), or cancel and close the
    window.

    This window can be used a modal window.
    In any case, the window will destroy itself.
    """

    def __init__(self, view, app, parent=None):
        BaseWindowController.__init__(self, view, app, parent=parent, actions={
            'delete_account': self.delete_account_action,
            'send_cancel': self.destroy
        })

    def show_modal(self):
        """Show the Windows and make it modal.

        Note: this method may block until the window is closed.
        """
        self.view.show_modal()

    @reduce_coroutine(safeguard=True)
    def delete_account_action(self, password):
        """User want to delete its account.

        Args:
            password (Text): user password. It's required as a protection for
                dangerous operations.
        """
        if not password:
            self.view.show_error(_('You must type your password account to '
                                   'change your passphrase'))
            return
        _logger.info('Send account deletion query ...')

        self.view.disable()
        try:
            yield self.app._user.delete_account(password)
        except HTTPUnauthorizedError:
            _logger.info('Account deletion rejected (bad password)')
            self._show_error(_('Invalid password!'))
            return
        except Exception as e:
            _logger.exception('Account deletion failed')
            self._show_error(_(u'An error occurred: \n%s' % err2unicode(e)))
            return
        self._destroy().then(self.app.disconnect)

    def _show_error(self, msg):
        self.view.enable()
        self.view.show_error(msg)

    def _destroy(self):
        self.destroyed.fire()
        return self.view.destroy()
